#include <netdb.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>

#include "response.h"
#include "scan.h"

extern pthread_mutex_t g_getservent;

static char *get_service_name(int port, char *proto)
{
	char *name = NULL;
	pthread_mutex_lock(&g_getservent);
	struct servent *servent = getservbyport(htons(port), proto);
	if (servent)
		name = strdup(servent->s_name);
	pthread_mutex_unlock(&g_getservent);
	return name;
}

void tcp_response(const struct tcphdr *tcphdr, const struct scan *data)
{
	if (data->type & SCAN_UDP) {
		dprintf(2,
		        "ft_nmap: error: received a TCP responses for an UDP "
		        "scan\n");
		return;
	}

	uint8_t type_index;
	for (uint8_t i = 0; i < NB_SCAN; i++)
	{
		if (data->type & (1 << i))
		{
			type_index = i;
			break;
		}	
	}

	if (data->responses->service == NULL)
		data->responses->service =
		    get_service_name(data->port_start, "tcp");
	if (data->type & SCAN_SYN) {
		if (tcphdr->ack == 1 && tcphdr->syn == 1)
			data->responses->states[type_index] = OPENED;
		else if (tcphdr->ack == 1 && tcphdr->rst == 1)
			data->responses->states[type_index] = CLOSED;
	} else if (data->type & SCAN_ACK && tcphdr->rst == 1)
		data->responses->states[type_index] = UNFILTERED;
	else if (data->type & SCAN_NULL && tcphdr->rst == 1)
		data->responses->states[type_index] = CLOSED;
	else if (data->type & SCAN_FIN && tcphdr->rst == 1)
		data->responses->states[type_index] = CLOSED;
	else if (data->type & SCAN_XMAS && tcphdr->rst == 1)
		data->responses->states[type_index] = CLOSED;
}

void udp_response(const struct udphdr *udphdr, const struct scan *data)
{
	(void)udphdr;
	if (data->type != SCAN_UDP) {
		dprintf(2,
		        "ft_nmap: error: received an UDP responses for a TCP "
		        "scan\n");
		return;
	}
	if (data->responses->service == NULL)
		data->responses->service =
		    get_service_name(data->port_start, "udp");
	data->responses->states[SCAN_UDP] = OPENED;
}

void icmp_response(const struct icmphdr *icmphdr, const struct scan *data)
{
	uint8_t type_index;
	for (uint8_t i = 0; i < NB_SCAN; i++)
	{
		if (data->type & (1 << i))
		{
			type_index = i;
			break;
		}
	}

	if (data->responses->service == NULL)
	{
		data->responses->service = get_service_name(data->port_start, "udp");
		if (data->responses->service == NULL)
			data->responses->service =
				get_service_name(data->port_start, "tcp");
	}
	if (data->type & SCAN_SYN && icmphdr->type == 3)
		data->responses->states[type_index] = FILTERED;
	else if (data->type & SCAN_ACK && icmphdr->type == 3)
		data->responses->states[type_index] = FILTERED;
	else if (data->type & SCAN_NULL && icmphdr->type == 3)
		data->responses->states[type_index] = FILTERED;
	else if (data->type & SCAN_FIN && icmphdr->type == 3)
		data->responses->states[type_index] = FILTERED;
	else if (data->type & SCAN_XMAS && icmphdr->type == 3)
		data->responses->states[type_index] = FILTERED;
	else if (data->type & SCAN_UDP && icmphdr->type == 3 && icmphdr->code == 3)
		data->responses->states[type_index] = CLOSED;
	else if (data->type & SCAN_UDP && icmphdr->type == 3 && icmphdr->code != 3)
		data->responses->states[type_index] = FILTERED;
}

void no_response(const struct scan *data)
{
	uint8_t type_index;
	for (uint8_t i = 0; i < NB_SCAN; i++)
	{
		if (data->type & (1 << i))
		{
			type_index = i;
			break;
		}	
	}

	if (data->responses->service == NULL)
	{
		data->responses->service = get_service_name(data->port_start, "udp");
		if (data->responses->service == NULL)
			data->responses->service =
				get_service_name(data->port_start, "tcp");
	}
	if (data->type & SCAN_SYN)
		data->responses->states[type_index] = FILTERED;
	else if (data->type & SCAN_ACK)
		data->responses->states[type_index] = FILTERED;
	else if (data->type & SCAN_NULL)
		data->responses->states[type_index] = OPENFILTERED;
	else if (data->type & SCAN_FIN)
		data->responses->states[type_index] = OPENFILTERED;
	else if (data->type & SCAN_XMAS)
		data->responses->states[type_index] = OPENFILTERED;
	else if (data->type & SCAN_UDP)
		data->responses->states[type_index] = OPENFILTERED;
}
